*******************************************************************************
* Project:	ultraSTE
* Module:	3d Objects
* Code:		ultra
* Created:	20.05.2021
* Notes:	
* Modified:
* Changes:
*
* (c) 2021 cream
*******************************************************************************

US3dGetVisibleFaces:
	move	#US3dVisibleFacesR,USRegSelect
	move	USRegData,d0
	rts

*******************************************************************************
US3dSetObjMatrix:
; a0 matrix (3x3 .w)
; d0 object (0-f)
	movem.l	d0/a0/a1,-(a7)
	and	#$f,d0
	add	#US3dWrObj1MatrixR,d0
	move	d0,USRegSelect
	US3dWriteMatrix		* a0 data 
	movem.l	(a7)+,d0/a0/a1
	rts

*******************************************************************************
US3dSetObjPos:
; a0 vec (3 x .w)
; d0 object (0-f)
	movem.l	d0/a0/a1,-(a7)
	and	#$f,d0
	add	#US3dWrObj1PosR,d0
	move	d0,USRegSelect
	US3dWriteVec	* a0 data
	movem.l	(a7)+,d0/a0/a1
	rts

*******************************************************************************
US3dSetObjRotationOffset:
; a0 vec (3 x .w)
; d0 object (0-f)
	movem.l	d0/a0/a1,-(a7)
	and	#$f,d0
	add	#US3dWrObj1RotOfsR,d0
	move	d0,USRegSelect
	US3dWriteVec	* a0 data
	movem.l	(a7)+,d0/a0/a1
	rts

*******************************************************************************
US3dSendObjects:
	moveq	#0,d0
	bra.s	US3dSendObjectsInt

US3dSendObjectsNT:
	moveq	#1,d0
	bra.s	US3dSendObjectsInt

US3dFaceFmt: dc.w 0
US3dSendObjectsInt:

; sends the object data (points, envmap points and facelist )
;
; a0 object list ptr to 16 object datas (see below)
; d0 0 = old face format ( p1,p2,p3,0 )
;    1 = new face format ( p1,p2,p3,0, vis, col, user1, user2 )
;
; objects are only send once for an effect to the fpga to save transfer time
; maximum size for the objects is currently 12kb all data must fit into this space!
;

; format:
; for each object
; pos
; 0 	dc.l <ptr to points>
; 4 	dc.l <ptr to enmap points (must be the same line the upper points)>
; 8		dc.l <prt to face list>
; 12	dc.w object positon x
; 14	dc.w object positon y
; 16	dc.w object positon z
; 18 	...
;
; use 0 for unused objects. also if your object does not
; have for example envmap points you can set it to 0
	move	d0, US3dFaceFmt
	move.l	a0,-(a7)
	lea	USRegData.w,a4
	lea	USRegSelect.w,a5

	; first calc the positions in result memory
	; and set object origins and rotation matrix (identity)
	moveq	#16-1,d0
	moveq	#0,d1		; global point count
	moveq	#0,d2		; global face count
	move	#0,d3		; obj index
.loop:

	move.l	US3dObjListPointOffset(a0),a2	;add points
	cmp.l	#0,a2
	beq.s	.notUsed
	add	(a2),d1
	addq	#1,d1

	;object has points so set it's origin
	lea	US3dObjListOriginX(a0),a2

	;obj position
	move	d3,d4
	add	#US3dWrObj1PosR,d4
	move	d4,(a5)	;USRegSelect
	move	(a2)+,(a4)	;USRegData
	move	(a2)+,(a4)	;USRegData
	move	(a2)+,(a4)	;USRegData

	;clear obj rotation offset
	move	d3,d4
	add	#US3dWrObj1RotOfsR,d4
	move	d4,(a5)	;USRegSelect
	clr.l	(a4)
	clr	(a4)

	; write identity matrix for the object
	move	d3,d4
	add	#US3dWrObj1MatrixR,d4
	move	d4,(a5)	;USRegSelect
	lea	US3dIdentityMat(pc),a3
	rept	3*3
	move	(a3)+,(a4)	;USRegData
	endr

;	exg.l	a3,a0
;	US3dWriteMatrix		; a0 data 
;	exg.l	a3,a0
.notUsed:

	move.l	US3dObjListFaceOffset(a0),a2	;add faces
	cmp.l	#0,a2
	beq.s	.notUsed2
	add	(a2),d2
	addq	#1,d2
.notUsed2:
	lea	US3dObjListEntryLen(a0),a0

	addq	#1,d3
	dbf	d0,.loop
	

	; fpga result memory organisation
	; points (always at 0)
	; envmap points
	; face list
	; face sort list

	;calc envmap start
	mulu	#3,d1		; 3*2=6 internal fpga mem is word based -> 3
	move	#US3dEnvMemLocR,(a5)	;USRegSelect
	move	d1,(a4)	;USRegData
	
	;calc face start
	add	d1,d1		; point + envmap points
	move	#US3dFaceMemLocR,(a5)	;USRegSelect
	move	d1,(a4)	;USRegData

	;calc face list start
	lsl	#3,d2		; old format 2 new 3
	add	d2,d1
	move	#US3dFaceSortMemLocR,(a5)	;USRegSelect
	move	d1,(a4)	;USRegData
	

	move.l	(a7)+,a0

	; send object datas
	move	#US3dWrObjectDataR,(a5)	;USRegSelect

	lea	US3dObjList(pc),a1
	move	#32-1,d7
.clrObjList:
	clr.l	(a1)+
	dbf	d7,.clrObjList

	
	lea	US3dObjList(pc),a1
	moveq	#16-1,d0
	moveq	#0,d7		; memory offset
.loop2:
	; points
	move.l	US3dObjListPointOffset(a0),a2
	cmp.l	#0,a2
	beq.s	.noPoints
	move	(a2)+,d1
	move	d1,d2

	addq	#1,d2
	move	d2,(a4)	;USRegData
.sendPointLoop:
	move	(a2)+,d5
;	neg	d5
	move	d5,(a4)	;USRegData
	move	(a2)+,d5
;	neg	d5
	move	d5,(a4)	;USRegData
	move	(a2)+,(a4)	;USRegData
	dbf	d1,.sendPointLoop


	move	d7,(a1)		;write points offset
	add	#256+64,(a1)

	mulu	#3,d2
	add	d2,d7
	add	#1,d7		;count value
.noPoints:

	; env points
	move.l	US3dObjListEnvOffset(a0),a2
	cmp.l	#0,a2
	beq.s	.noEnv
	move	(a2)+,d1
	move	d1,d2

	addq	#1,d2
	move	d2,(a4)	;USRegData
.sendEnvLoop:
	move	(a2)+,(a4)	;USRegData
	move	(a2)+,(a4)	;USRegData
	move	(a2)+,(a4)	;USRegData
	dbf	d1,.sendEnvLoop

	move	d7,2(a1)	;write env points offet
	add	#256+64,2(a1)
	mulu	#3,d2
	add	d2,d7
	add	#1,d7		;count value
.noEnv:

	; faces
	move.l	US3dObjListFaceOffset(a0),a2
	cmp.l	#0,a2
	beq.s	.noFaces
	move	(a2)+,d1
	move	d1,d2

	addq	#1,d2
	move	d2,(a4)	;USRegData
	tst	US3dFaceFmt
	beq.s	.doOldFaceFormat

	moveq	#0,d6
.sendFacesLoop2:
	move	(a2)+,(a4)	;USRegData		;point1
	move	(a2)+,(a4)	;USRegData		;point2
	move	(a2)+,(a4)	;USRegData		;point3
	move	(a2)+,(a4)	;USRegData		;dummy facez
	move	(a2)+,(a4)	;USRegData		;face visibility
	move	(a2)+,(a4)	;USRegData		;face color
	move	(a2)+,(a4)	;USRegData		;face user1
	move	(a2)+,(a4)	;USRegData		;face user2
	dbf	d1,.sendFacesLoop2
	bra.s	.wasNew

.doOldFaceFormat:
	moveq	#0,d6
.sendFacesLoop:
	move	(a2)+,(a4)	;USRegData		;point1
	move	(a2)+,(a4)	;USRegData		;point2
	move	(a2)+,(a4)	;USRegData		;point3
	move	(a2)+,(a4)	;USRegData		;dummy facez
	clr	(a4)	;USRegData		;face visibility
;//	clr	(a4)	;USRegData		;face color
	move	d6,(a4)	;USRegData		;face color
	eor	#2,d6
	move	d0,(a4)
	;clr	(a4)	;USRegData		;face user1
	clr	(a4)	;USRegData		;face user2
	dbf	d1,.sendFacesLoop

.wasNew:
	move	d7,4(a1)		;write face offset
	add	#256+64,4(a1)

	lsl	#3,d2
	add	d2,d7
	add	#1,d7		;count value
.noFaces:

	lea	US3dObjListEntryLen(a0),a0
	lea	4*2(a1),a1

	dbf	d0,.loop2

	lea	US3dObjList(pc),a0
	move	#US3dWrObjectListR,(a5)	;USRegSelect
	move	#(US3dObjListEntryFPGALen*16/2)-1,d0
.sendObjectList:
	move	(a0)+,(a4)	;USRegData
	dbf	d0, .sendObjectList
 	rts
	
;	bsr		US3dStartCalc
;	bsr		US3dWaitBusy

	lea	$100000,a0
	bsr	US3dDumpObjectMemory
	lea	$110000,a0
	bsr	US3dDumpResultMemory

	lea	US3dObjList(pc),a0
	move.l	#$123546,d0
	illegal
	rts

US3dDumpObjectMemory:
; dumps the internal object memory
; just for debugging
;
; a0 -> buffer to write to

	move	#USObjMemRdPos,USRegSelect
	move	#0,USRegData

	move	#USObjMemData,USRegSelect
	move	#8192-1,d0
.readLoop:
	move	USRegData,(a0)+
	dbf	d0,.readLoop
	rts

US3dDumpResultMemory:
; dumps the internal result memory
; just for debugging
;
; a0 -> buffer to write to

	move	#US3dRdVertex,USRegSelect

	move	#(8192/2)-1,d0
.readLoop:
	move.l	US3dResultData,(a0)+
	dbf	d0,.readLoop
	rts

US3dIdentityMat:
	dc.w	$4000,$0000,$0000
	dc.w	$0000,$4000,$0000
	dc.w	$0000,$0000,$4000
US3dObjList:
	ds.w	16*4
